// Code for Light

// Libraries 
#include <WiFi.h>
#include <PubSubClient.h>   // For MQTT 
#include <RBDdimmerESP32.h> // for dimmer => download here = https://github.com/RobotDynOfficial/Lib-RBD-Dimmer-for-ESP32
// For dimmer code, see here => http://www.lesmartomation.com/ac-light-dimmer-module-with-fade-effect/

// Free Smart Home parameters 
    const char* ssid     = "<your-wifi-name>";
    const char* password = "<your-wifi-password>";

    const char* server_ip = "<ip-server-mqtt>";   // ip address of the raspberry pi server MQTT
    String ESP32_Number = "2";    // Each ESP32 module has to get a number
    String MQTT_sub_topic1 = "ESP32/LightOutput/" + ESP32_Number; // sub topic name 
    String MQTT_reply_topic = "ESP32/LightOutput/response/" + ESP32_Number;

// defines pins numbers
    const int light_A_Pin = 23;    // pin relay
    const int light_B_Pin= 22;
    const int light_C_Pin= 32;
    const int light_D_Pin= 33;
    
    const int PWM_A_Pin=25;       // pin PWM output for dimmer
    const int PWM_B_Pin=26;  
    const int PWM_C_Pin=27;
    const int PWM_D_Pin=14;
    
    const int PWM_A_ZC_Pin=19;    // For dimmer, it's GPIO for Zero-Cross, don't use pins 28(GPIO9), 29(GPIO10), 30(GPIO11), 31(GPIO6), 33(GPIO8)
    const int PWM_B_ZC_Pin=18;
    const int PWM_C_ZC_Pin=17;
    const int PWM_D_ZC_Pin=16;

// Initialisation for dimmer
    dimmerLampESP32 dimmer_A(PWM_A_Pin, PWM_A_ZC_Pin); //initialase port for dimmer_A
    dimmerLampESP32 dimmer_B(PWM_B_Pin, PWM_B_ZC_Pin); //initialase port for dimmer_B
    dimmerLampESP32 dimmer_C(PWM_C_Pin, PWM_C_ZC_Pin); //initialase port for dimmer_C
    dimmerLampESP32 dimmer_D(PWM_D_Pin, PWM_D_ZC_Pin); //initialase port for dimmer_D
    int dutyCycle_A =0;
    int dutyCycle_B =0;
    int dutyCycle_C =0;
    int dutyCycle_D =0;




// For wifi conection 
    WiFiServer server(80);    // Object containing the web server. Communication by the port 80.
    String header; // Variable to store the HTTP request

// For MQTT server 
    // MQTT Broker IP address, example of my raspberry py
    const char* mqtt_server = server_ip;
    WiFiClient espClient;
    PubSubClient client(espClient);
    long lastMsg = 0;
    char msg[50];
    int value = 0;



void setup_wifi(){
    // Wifi init   
       Serial.begin(115200);
       delay(10); 
       Serial.print("Connecting to ");
       Serial.println(ssid);
       WiFi.begin(ssid, password);
            
       while (WiFi.status() != WL_CONNECTED) {   // stay here while it's not connected. 
         delay(500);
         Serial.print(".");
       }
            
    // Print local IP address and start web server
       Serial.println("");
       Serial.println("WiFi connected.");;
       //server.begin(); // starting server 
       //Serial.print("Server started. Open a web browser with the ip address  ");
       Serial.println("IP address: ");
       Serial.println(WiFi.localIP());
       Serial.println("");
}


String getValue(String data, char separator, int index) // function used to separate a string "data" with a separator
{
    int found = 0;
    int strIndex[] = { 0, -1 };
    int maxIndex = data.length() - 1;

    for (int i = 0; i <= maxIndex && found <= index; i++) {
        if (data.charAt(i) == separator || i == maxIndex) {
            found++;
            strIndex[0] = strIndex[1] + 1;
            strIndex[1] = (i == maxIndex) ? i+1 : i;
        }
    }
    return found > index ? data.substring(strIndex[0], strIndex[1]) : "";
}


void publishReply(String tempTopicName, String tempReply ) {
  tempReply = "Module_" + ESP32_Number + "/" + tempReply;
  client.publish((char*) tempTopicName.c_str(), (char*) tempReply.c_str() );
}

void callback(char* topic, byte* message, unsigned int length) {    //Fonction for MQTT reception
      Serial.print("Message arrived on topic: ");
      Serial.print(topic);
      Serial.print(". Message: ");
      String messageTemp;
      
      for (int i = 0; i < length; i++) {
        Serial.print((char)message[i]);
        messageTemp += (char)message[i];
      }
      Serial.println();



   

        // Include here the code for GPIO management with ESP32 as subscriber
    
        // If a message is received on the topic esp32/LEDoutput
            
            if (String(topic) == MQTT_sub_topic1)  // Name of the topic
            {  
                    // message is "output_port/name/on or off or pwm in %
                    String LightPort = getValue(messageTemp, '/', 0);
                    String LightState = getValue(messageTemp, '/', 2);
                    
      
                  // ---------- Light TOR ( A/B/C/D ) -----------
                  
                          if (LightPort== "A"){
                                if (LightState == "on"){  // switch on
                                    digitalWrite(light_A_Pin, LOW); // relay commanded with Low signal 
                                    Serial.println("Light_A_Pin ---- ON");
                                    publishReply(MQTT_reply_topic, LightPort + "/" + LightState);
                                 }else{ // switch off
                                    digitalWrite(light_A_Pin, HIGH);
                                    Serial.println("Light_A_Pin ---- OFF");
                                    publishReply(MQTT_reply_topic, LightPort + "/" + LightState);
                                 }
                          }else if (LightPort== "B"){
                                if (LightState == "on"){  // switch on
                                    digitalWrite(light_B_Pin, LOW); // relay commanded with Low signal 
                                    Serial.println("Light_B_Pin ---- ON");
                                    publishReply(MQTT_reply_topic, LightPort + "/" + LightState);
                                 }else{ // switch off
                                    digitalWrite(light_B_Pin, HIGH);
                                    Serial.println("Light_B_Pin ---- OFF");
                                    publishReply(MQTT_reply_topic, LightPort + "/" + LightState);
                                 }
                          }else if (LightPort== "C"){
                                if (LightState == "on"){  // switch on
                                    digitalWrite(light_C_Pin, LOW); // relay commanded with Low signal 
                                    Serial.println("Light_C_Pin ---- ON");
                                    publishReply(MQTT_reply_topic, LightPort + "/" + LightState);
                                 }else{ // switch off
                                    digitalWrite(light_C_Pin, HIGH);
                                    Serial.println("Light_C_Pin ---- OFF");
                                    publishReply(MQTT_reply_topic, LightPort + "/" + LightState);
                                 }
                          }else if (LightPort== "D"){
                                if (LightState == "on"){  // switch on
                                    digitalWrite(light_D_Pin, LOW); // relay commanded with Low signal 
                                    Serial.println("Light_D_Pin ---- ON");
                                    publishReply(MQTT_reply_topic, LightPort + "/" + LightState);
                                 }else{ // switch off
                                    digitalWrite(light_D_Pin, HIGH);
                                    Serial.println("Light_D_Pin ---- OFF");
                                    publishReply(MQTT_reply_topic, LightPort + "/" + LightState);
                                 }
                          }
                          

                    // ---------- Light PWM ( PWM_A/PWM_B/PWM_C/PWM_D ) -----------


                          else if (LightPort== "PWM_A"){

                                // PWM A activate
                                    if (LightState == "on" || LightState == "100"){
                                      dutyCycle_A = 100;
                                      dimmer_A.setPower(99); // setPower(0-100%)
                                    }else if (LightState == "off" || LightState == "0"){
                                      dutyCycle_A = 0;
                                      dimmer_A.setPower(1); // setPower(0-100%)
                                    } else {
                                      dutyCycle_A = LightState.toInt();
                                      dimmer_A.setPower(dutyCycle_A); // setPower(0-100%)
                                    }
                                    Serial.print("PWM_A_Pin at ");
                                    Serial.print(dutyCycle_A);
                                    Serial.println(" %");
                                    publishReply(MQTT_reply_topic, LightPort + "/" + dutyCycle_A);
                          
                          } else if (LightPort== "PWM_B"){

                                // PWM B activate
                                    if (LightState == "on" || LightState == "100"){
                                      dutyCycle_B = 100;
                                      dimmer_B.setPower(99); // setPower(0-100%)
                                    }else if (LightState == "off" || LightState == "0"){
                                      dutyCycle_B = 0;
                                      dimmer_B.setPower(1); // setPower(0-100%)
                                    } else {
                                      dutyCycle_B = LightState.toInt();
                                      dimmer_B.setPower(dutyCycle_B); // setPower(0-100%)
                                    }
                                    Serial.print("PWM_B_Pin at ");
                                    Serial.print(dutyCycle_B);
                                    Serial.println(" %");
                                    publishReply(MQTT_reply_topic, LightPort + "/" + dutyCycle_B);
                                      
                          } else if (LightPort== "PWM_C"){

                                // PWM C activate
                                    if (LightState == "on" || LightState == "100"){
                                      dutyCycle_C = 100;
                                      dimmer_C.setPower(99); // setPower(0-100%)
                                    }else if (LightState == "off" || LightState == "0"){
                                      dutyCycle_C = 0;
                                      dimmer_C.setPower(1); // setPower(0-100%)
                                    } else {
                                      dutyCycle_C = LightState.toInt();
                                      dimmer_C.setPower(dutyCycle_C); // setPower(0-100%)
                                    }
                                    
                                    Serial.print("PWM_C_Pin at ");
                                    Serial.print(dutyCycle_C);
                                    Serial.println(" %");
                                    publishReply(MQTT_reply_topic, LightPort + "/" + dutyCycle_C);
                                      
                          } else if (LightPort== "PWM_D"){

                                // PWM D activate
                                    if (LightState == "on" || LightState == "100"){
                                      dutyCycle_D = 100;
                                      dimmer_D.setPower(99); // setPower(0-100%)
                                    }else if (LightState == "off" || LightState == "0"){
                                      dutyCycle_D = 0;
                                      dimmer_D.setPower(1); // setPower(0-100%)
                                    } else {
                                      dutyCycle_D = LightState.toInt();
                                      dimmer_D.setPower(dutyCycle_D); // setPower(0-100%)
                                    }
                                    
                                    Serial.print("PWM_D_Pin at ");
                                    Serial.print(dutyCycle_D);
                                    Serial.println(" %");
                                    publishReply(MQTT_reply_topic, LightPort + "/" + dutyCycle_D);
                                      
                          } else {

                                Serial.println();
                                Serial.println("Error. Wrong command from server");
                                Serial.println();
                                publishReply(MQTT_reply_topic, LightPort + " ----- ERROR -----");
                          }
                  } 
}

void reconnect() {    // Function for MQTT server connection
    // Loop until we're reconnected
    while (!client.connected()) {
        Serial.print("Attempting MQTT connection...");
        // Attempt to connect
        String ClientName = "ESP32ClientLight" + ESP32_Number; 
        if (client.connect("ClientName")) { // ESP32 name dor connection to MQTT server
          Serial.println("connected");
  
  
          
          // Write here topics to subscribe
          client.subscribe((char*) MQTT_sub_topic1.c_str());
          Serial.println("Client name : " + ClientName);
          Serial.println("Listening on topic : " + MQTT_sub_topic1);
          Serial.println();



        } else {
          Serial.print("failed, rc=");
          Serial.print(client.state());
          Serial.println(" try again in 5 seconds");
          // Wait 5 seconds before retrying
          delay(5000);
      }
    }
}



void setup() {
        setup_wifi();
      // Server MQTT
      client.setServer(mqtt_server, 1883);
      client.setCallback(callback);
     
      // Output
      pinMode(light_A_Pin, OUTPUT);
      pinMode(light_B_Pin, OUTPUT); 
      pinMode(light_C_Pin, OUTPUT); 
      pinMode(light_D_Pin, OUTPUT); 
      digitalWrite(light_A_Pin, HIGH); // To init relay OFF
      digitalWrite(light_B_Pin, HIGH);
      digitalWrite(light_C_Pin, HIGH);
      digitalWrite(light_D_Pin, HIGH);

      // For dimmer
      dimmer_A.begin(NORMAL_MODE, ON); //dimmer initialisation: name.begin(MODE, STATE)
      dimmer_B.begin(NORMAL_MODE, ON); //dimmer initialisation: name.begin(MODE, STATE) 
      dimmer_C.begin(NORMAL_MODE, ON); //dimmer initialisation: name.begin(MODE, STATE) 
      dimmer_D.begin(NORMAL_MODE, ON); //dimmer initialisation: name.begin(MODE, STATE) 
}

void loop() {
  
    if (!client.connected()) {
        reconnect();
      }
      client.loop();  // Loop function for topic listening 
    
      long now = millis();
      if (now - lastMsg > 5000) {
        lastMsg = now;  
      }

    
      // include here the code for sensors


     
      // include here messages to publish to MQTT server
       //   String message = "Trying to send a message";
       //   client.publish("esp32/Distance", (char*) message.c_str() );
}
