
// Import the different library

#include <WiFi.h>
#include <PubSubClient.h>   
#include <DHT.h>

#define DHTPIN 15     // what digital pin we're connected to

#define DHTTYPE DHT22  //type of DHT module
DHT dht(DHTPIN, DHTTYPE);


// defines pins numbers
const int BO_trigger_1 = 5;
const int BI_echo_1 = 4;


// Parameters for MQTT serve/wifi connection
    const char* ssid     = "<your-wifi-name>";
    const char* password = "<your-wifi-password>";


    IPAddress local_ip(192,168,1,103);
    IPAddress gateway(192,168,0,1);
    IPAddress subnet(255,255,255,0);
    WiFiServer server(80);    // On créé une instance, un objet qui contiendra le serveur web. On communiquera avec lui sur le port 80, le port classique des requêtes HTTP utilisées par les navigateurs internet.
    String header; // Variable to store the HTTP request

    // MQTT Broker IP address, example of my raspberry py
    const char* mqtt_server = "<ip-server-mqtt>";   // l'adresse complete est 192.168.0.34:1883
    //const char* mqtt_server = "YOUR_MQTT_BROKER_IP_ADDRESS";

    WiFiClient espClient;
    PubSubClient client(espClient);
    long lastMsg = 0;
    char msg[50];
    int value = 0;



void setup_wifi(){
    // WIfi init 
       Serial.begin(115200); // Init serial monitor
       delay(10); 
       Serial.print("Connecting to ");
       Serial.println(ssid);
       WiFi.softAP(ssid, password);
       WiFi.softAPConfig(local_ip, gateway, subnet);
       delay(100);
            
            
    // Print local IP address and start web server
       Serial.println("");
       Serial.println("WiFi connected.");;
       //server.begin(); // démare le serveur 
       //Serial.print("Server started. Open a web browser with the ip address  ");
       Serial.println("IP address: ");
       Serial.println(WiFi.localIP());
       Serial.println("");

}


//Function which measure the resistor of the thermistor and convert to a temperature
float thermistorMesure(int PinThermistor,float R1){
  int Vo;
  float logR2, R2, T, Tc;
  float c1 = 1.009249522e-03, c2 = 2.378405444e-04, c3 = 2.019202697e-07;

  Vo = analogRead(PinThermistor);
  R2 = R1 * (4095.0 / (float)Vo - 1.0);
  logR2 = log(R2);
  T = (1.0 / (c1 + c2*logR2 + c3*logR2*logR2*logR2));
  Tc = T - 273.15;
  return Tc;  
  
}

//Function which permits to convert the measure of the ultrasonic sensor into a distance
float ultrasonMesure(int PinTrigger, int PinEcho){  
      float distance = 0.0;
      long duration = 0;
      
       // Clears the trigPin
      digitalWrite(PinTrigger, LOW);
      delayMicroseconds(2);
      // Sets the PinTrigger on HIGH state for 10 micro seconds
      digitalWrite(PinTrigger, HIGH);
      delayMicroseconds(10);
      digitalWrite(PinTrigger, LOW);
      // Reads the PinEcho, returns the sound wave travel time in microseconds
      duration = pulseIn(PinEcho, HIGH,32760);
      // Calculating the distance
      distance= (duration*0.034/2);
    
      return distance;
}


void callback(char* topic, byte* message, unsigned int length) {    
      Serial.print("Message arrived on topic: ");
      Serial.print(topic);
      Serial.print(". Message: ");
      String messageTemp;
      
      for (int i = 0; i < length; i++) {
        Serial.print((char)message[i]);
        messageTemp += (char)message[i];
      }
      Serial.println();
}



void reconnect() {  
    // Loop until we're reconnected
    while (!client.connected()) {
      Serial.print("Attempting MQTT connection...");
      // Attempt to connect
      if (client.connect("ESP32Client")) {
        Serial.println("connected");
        // Subscribe
        client.subscribe("outside_station/led");
      } else {
        Serial.print("failed, rc=");
        Serial.print(client.state());
        Serial.println(" try again in 5 seconds");
        // Wait 5 seconds before retrying
        delay(5000);
      }
    }
}





//Function to setup the program
void setup() {
  
      setup_wifi();
      // MQTT server
      client.setServer(mqtt_server, 1883);
      client.setCallback(callback);
     
      //Declaratio of input/output pins
      pinMode(BO_trigger_1, OUTPUT);  
      pinMode(BI_echo_1, INPUT); 
      dht.begin();


}

void loop() {
      Serial.println(WiFi.localIP());
      if (!client.connected()) {
        reconnect();
      }
      //Function loop to read never stop the mqtt values
      client.loop();  
    
      long now = millis();
      if (now - lastMsg > 5000) {
        lastMsg = now;  
      }

          //Get the process values and convert them into strings
          String distance = String(ultrasonMesure(BO_trigger_1,BI_echo_1));    
          String temperature = String(dht.readTemperature());
          String humidity = String(dht.readHumidity());

          
          Serial.print("Distance : ");
          Serial.println(distance);
          Serial.print("Temperature : ");
          Serial.println(temperature);
          Serial.print("Humidite : ");
          Serial.println(humidity);
 

          //Send the process values to the mqtt server
          client.publish("outside_station/distance", (char*) distance.c_str() );
          client.publish("outside_station/temperature", (char*) temperature.c_str() );
          client.publish("outside_station/humidity", (char*) humidity.c_str() );
          delay (1000);
}
